/* Class.c -- implementation of class descriptor objects

	THIS SOFTWARE FITS THE DESCRIPTION IN THE U.S. COPYRIGHT ACT OF A
	"UNITED STATES GOVERNMENT WORK".  IT WAS WRITTEN AS A PART OF THE
	AUTHOR'S OFFICIAL DUTIES AS A GOVERNMENT EMPLOYEE.  THIS MEANS IT
	CANNOT BE COPYRIGHTED.  THIS SOFTWARE IS FREELY AVAILABLE TO THE
	PUBLIC FOR USE WITHOUT A COPYRIGHT NOTICE, AND THERE ARE NO
	RESTRICTIONS ON ITS USE, NOW OR SUBSEQUENTLY.

Author:
	K. E. Gorlen
	Bg. 12A, Rm. 2017
	Computer Systems Laboratory
	Division of Computer Research and Technology
	National Institutes of Health
	Bethesda, Maryland 20892
	Phone: (301) 496-5363
	uucp: {decvax!}seismo!elsie!cecil!keith
	September, 1985

Function:
	
Functions pertaining to class Class.  Most of these are defined inline
in object.hxx.

Modification History:

1-Nov-86	K. E. Gorlen

1.  Added classNumber.

06-Oct-86	S. M. Orlow

1.  Added binary I/O constructor, storer, and reader functions

29-Apr-86	K. E. Gorlen

1.  Compute the class_version number from the version numbers supplied
in the DEFINE_CLASS macros of each class and all its base classes.

12-Apr-86	K. E. Gorlen

1.  Add class initor2 function, called after all OOPS classes have been
initialized and the classDictionary has been built.

*/

#include "Object.hxx"
#include "String.hxx"
#include "Dictionary.hxx"
#include "Assoc.hxx"

#define	THIS	Class
#define	BASE	Object
/*DEFINE_CLASS(Class,Object,1,NULL,NULL);	// cfront 1.2.1 bug */
extern const int OOPS_RDUNKCLASS;
overload Class_reader;
static void Class_reader(istream& /*strm*/, Object& /*where*/)
{
	setOOPSerror(OOPS_RDUNKCLASS,DEFAULT,"Class");
}
static void Class_reader(fileDescTy& /*fd*/, Object& /*where*/)
{
	setOOPSerror(OOPS_RDUNKCLASS,DEFAULT,"Class");
}
Class class_Class = Class( class_Object, "Class", 1, sizeof(Class),Class_reader, Class_reader, NULL, NULL);

const Class* Class::isA()	{ return &class_Class; }

// head of the list of all Classes, built by the static constructor of class Class 
Class* allClasses =0;

Dictionary classDictionary;	// Dictionary of all Classes 

Class::Class(const Class& super, const char* name, UNSIGNED version, UNSIGNED size, void (*reader)(istream&,Object&), void (*binreader)(fileDescTy&,Object&), initorTy initor1, initorTy initor2)
{
	superClass = &super;
	nextClass = allClasses; allClasses = this;
	class_name = name;
	def_version = version;
	class_version = 0;
	inst_size = size;
	inst_reader = reader;
        inst_binreader = binreader;
	class_number = 0;
	class_initor2 = initor2;
//	if (initor1 != 0) initor1(*this);	MASSCOMP cc bug
	register initorTy init = initor1;
	if (init != 0) init(*this);
}

int Class::compare(const Object& ob)	// compare Class names 
{
	assertArgSpecies(ob,class_Class,"compare");
	return strcmp(className(),((Class*)&ob)->className());
}

obid Class::deepCopy()
{
	shouldNotImplement("deepCopy");
	return 0;
}

UNSIGNED Class::hash()	{ return (UNSIGNED)this; }

bool Class::isEqual(const Object& ob)	{ return this == (Class*)&ob; }

void Class::printOn(ostream& strm)	{ strm << class_name; }

obid Class::shallowCopy()
{
	shouldNotImplement("shallowCopy");
	return 0;
}

UNSIGNED Class::size()	{ return inst_size; }
	
void OOPS__classInit()	// class Initialization
{
	register Class* clp = allClasses;	// calculate class version numbers
	while (clp != NULL) {
		register Class* clq = clp;
		while (clq != NULL) {
			clp->class_version = (clp->class_version << 4)
				^ (clp->class_version >> (8*sizeof(int)-4) & 0xF)
				^ clq->def_version;
			clq = clq->baseClass();
		}
		clp = clp->nextClass;
	}
	clp = allClasses;	// add classes to classDictionary
	while (clp != NULL) {
		classDictionary.add(*new Assoc(*new String(clp->className()), *clp));
		clp = clp->nextClass;
	}
	clp = allClasses;	// call the initor2 functions
	while (clp != NULL) {
		if (clp->class_initor2 != 0) (*clp->class_initor2)(*clp);
		clp = clp->nextClass;
	}
}

void Class::storer(ostream&) { shouldNotImplement("storer"); }

void Class::storer(fileDescTy&) { shouldNotImplement("storer"); }
